use super::Backup;
use serde::de::DeserializeOwned;
use serde::Serialize;
use std::fs::File;
use std::io;
use std::path::PathBuf;

pub struct Manager<T>
where
    T: Clone + Serialize + DeserializeOwned,
{
    backup_dir: PathBuf,
    config_dir: PathBuf,
    world_dir: PathBuf,
    default_metadata: T,
    chain_len: u64,
    chains_to_keep: u64,
    chains: Vec<Vec<Backup<T>>>,
}

impl<T> Manager<T>
where
    T: Clone + Serialize + DeserializeOwned,
{
    const METADATA_FILE: &str = "alex.json";

    /// Initialize a new instance of a `BackupManager`.
    pub fn new(
        backup_dir: PathBuf,
        config_dir: PathBuf,
        world_dir: PathBuf,
        metadata: T,
        chain_len: u64,
        chains_to_keep: u64,
    ) -> Self {
        Self {
            backup_dir,
            config_dir,
            world_dir,
            default_metadata: metadata,
            chain_len,
            chains_to_keep,
            chains: Vec::new(),
        }
    }

    /// Create a new backup with the expected type.
    pub fn create_backup(&mut self) -> io::Result<()> {
        let dirs = vec![
            (PathBuf::from("config"), self.config_dir.clone()),
            (PathBuf::from("worlds"), self.world_dir.clone()),
        ];

        // We start a new chain if the current chain is complete, or if there isn't a first chain
        // yet
        if let Some(current_chain) = self.chains.last() {
            let current_chain_len: u64 = current_chain.len().try_into().unwrap();

            if current_chain_len >= self.chain_len {
                self.chains.push(Vec::new());
            }
        } else {
            self.chains.push(Vec::new());
        }

        let current_chain = self.chains.last_mut().unwrap();

        let mut backup = if !current_chain.is_empty() {
            let previous_backup = current_chain.last().unwrap();
            let state = Backup::state(current_chain);

            Backup::create_from(state, previous_backup.start_time, &self.backup_dir, dirs)?
        } else {
            Backup::create(&self.backup_dir, dirs)?
        };

        backup.set_metadata(self.default_metadata.clone());

        current_chain.push(backup);

        self.save()?;

        Ok(())
    }

    /// Delete all backups associated with outdated chains, and forget those chains.
    pub fn remove_old_backups(&mut self) -> io::Result<()> {
        let chains_to_store: usize = self.chains_to_keep.try_into().unwrap();

        if chains_to_store < self.chains.len() {
            let mut remove_count: usize = self.chains.len() - chains_to_store;

            // We only count finished chains towards the list of stored chains
            let chain_len: usize = self.chain_len.try_into().unwrap();
            if self.chains.last().unwrap().len() < chain_len {
                remove_count -= 1;
            }

            for chain in self.chains.drain(..remove_count) {
                for backup in chain {
                    let path = Backup::path(&self.backup_dir, backup.start_time);
                    std::fs::remove_file(path)?;
                }
            }

            self.save()?;
        }

        Ok(())
    }

    /// Write the in-memory state to disk.
    pub fn save(&self) -> io::Result<()> {
        let json_file = File::create(self.backup_dir.join(Self::METADATA_FILE))?;
        serde_json::to_writer(json_file, &self.chains)?;

        Ok(())
    }

    /// Overwrite the in-memory state with the on-disk state.
    pub fn load(&mut self) -> io::Result<()> {
        let json_file = match File::open(self.backup_dir.join(Self::METADATA_FILE)) {
            Ok(f) => f,
            Err(e) => {
                // Don't error out if the file isn't there, it will be created when necessary
                if e.kind() == io::ErrorKind::NotFound {
                    self.chains = Vec::new();

                    return Ok(());
                } else {
                    return Err(e);
                }
            }
        };

        self.chains = serde_json::from_reader(json_file)?;

        Ok(())
    }
}
